import copy
from typing import List, Dict

import autobridge.Floorplan.Utilities as util
import autobridge.Floorplan as autobridge_floorplan
from autobridge.Device.DeviceManager import DeviceManager
from autobridge.HLSParser.tapa.DataflowGraphTapa import DataflowGraphTapa
from autobridge.HLSParser.tapa.ProgramJsonManager import ProgramJsonManager
from autobridge.Opt.DataflowGraph import Edge, Vertex
from autobridge.Opt.Slot import Slot
from autobridge.Opt.SlotManager import SlotManager
from autobridge.Route.global_route import ILPRouter
from autobridge.Route.latency_balancing import latency_balancing
from autobridge.util import *
from autobridge.analyze import analyze_result, analyze_input, is_device_supported

def annotate_floorplan(config: Dict) -> Dict:
  cli_logger = set_general_logger(config)

  print_start(config)

  init_logging(config)

  if not is_device_supported(config):
    return config

  analyze_input(config)

  board = DeviceManager(
      board_name=get_board_num(config),
      ddr_list=get_ddr_list(config),
      is_vitis_enabled=True,
  ).getBoard()
  program_json_manager = ProgramJsonManager(
      config['edges'],
      get_vertex_section(config),
      get_area_section(config),
  )
  graph = DataflowGraphTapa(program_json_manager)
  slot_manager = SlotManager(board)

  # which modules must be assigned to the same slot
  grouping_constraints: List[List[str]] = config.get('grouping_constraints', [])

  # process optional module pre-assignment constraints
  module_floorplan = config['floorplan_pre_assignments']
  pre_assignment = {}
  for region, module_group in module_floorplan.items():
    for mod_name in module_group:
      pre_assignment[mod_name] = region

  kwargs = get_floorplan_params(config)

  # generate floorplan
  # slot_list includes all possible slot regardless of whether it is empty
  v2s, slot_list = autobridge_floorplan.get_floorplan(
    graph,
    slot_manager,
    grouping_constraints,
    pre_assignment,
    **kwargs,
  )

  # if floorplan failed
  if v2s is None:
    cli_logger.critical('\n*** CRITICAL WARNING: AutoBridge fails to find a solution. Please refer to the log for details.\n')
    config['floorplan_status'] = 'FAILED'
    print_end()
    return config

  slot_to_usage = util.get_slot_utilization(v2s)

  # route the FIFO pipelines
  router = ILPRouter(
    list(graph.edges.values()),
    v2s,
    slot_to_usage,
    slot_list,
  )
  fifo_to_path: Dict[Edge, List[Slot]] = router.route_design()

  fifo_name_to_depth = latency_balancing(graph, fifo_to_path)

  annotated_config = get_annotated_config(v2s, fifo_to_path, slot_to_usage, fifo_name_to_depth, config)

  analyze_result(annotated_config)

  print_end()

  return annotated_config

def get_floorplan_params(config) -> Dict:
  kwargs = {}
  params = (
    'floorplan_strategy',
    'floorplan_opt_priority',
    'floorplan_opt_priority',
    'min_area_limit',
    'max_area_limit',
    'min_slr_width_limit',
    'max_slr_width_limit',
    'max_search_time',
  )
  for param in params:
    if param in config:
      kwargs[param] = config[param]

  # u280 only: get the hbm port list
  if config['part_num'].startswith('xcu280') and config.get('enable_hbm_binding_adjustment', False):
    hbm_port_v_name_list = []
    for v, props in config['vertices'].items():
      if props['category'] == 'PORT_VERTEX' and props['port_cat'] == 'HBM':
        hbm_port_v_name_list.append(v)
    kwargs['hbm_port_v_name_list'] = hbm_port_v_name_list

  return kwargs

def get_ddr_list(config) -> List[int]:
  """ extract which ddrs are enabled
  """
  ddr_set = set()
  for v_name, properties in config['vertices'].items():
    if properties['category'] == 'PORT_VERTEX':
      if properties['port_cat'] == 'DDR':
        ddr_set.add(properties['port_id'])
  return list(ddr_set)

def get_board_num(config) -> str:
  device_num = config['part_num']
  if device_num == 'xcu250-figd2104-2L-e':
    return 'U250'
  elif device_num == 'xcu280-fsvh2892-2L-e':
    return 'U280'
  else:
    raise NotImplementedError(f'unsupported device {device_num}')

def get_vertex_section(config) -> Dict[str, str]:
  return {v_name: properties['module'] for v_name, properties in config['vertices'].items()}

def get_area_section(config) -> Dict[str, Dict[str, int]]:
  return {properties['module']: properties['area'] for v_name, properties in config['vertices'].items()}

def get_annotated_config(
    v2s: Dict[Vertex, Slot],
    fifo_to_path: Dict[Edge, List[Slot]],
    slot_to_usage: Dict[Slot, Dict[str, float]],
    fifo_name_to_depth: Dict[str, int],
    config_orig: Dict,
) -> Dict:
  config = copy.deepcopy(config_orig)
  for v, s in v2s.items():
    config['vertices'][v.name]['floorplan_region'] = s.getRTLModuleName()
    config['vertices'][v.name]['SLR'] = s.getSLR()

  # which slots are not empty
  used_slots_list = list(v2s.values()) + [s for path in fifo_to_path.values() for s in path ]
  used_slots = set(used_slots_list)

  config['floorplan_region_pblock_tcl'] = {s.getRTLModuleName(): s.pblock_tcl for s in used_slots}

  for fifo, path in fifo_to_path.items():
    config['edges'][fifo.name]['path'] = [s.name for s in path]

  config['slot_resource_usage'] = {s.getRTLModuleName(): usage for s, usage in slot_to_usage.items()}
  config['floorplan_status'] = 'SUCCEED'

  # update the edge depth
  for fifo_name, depth in fifo_name_to_depth.items():
    config['edges'][fifo_name]['adjusted_depth'] = depth

  # record important floorplan metrics
  config['actual_slr_width_usage'] = util.get_actual_slr_crossing_limit(v2s)
  config['actual_area_usage'] = util.get_actual_area_limit(v2s)

  # record the new HBM binding
  if any(s.isHalfSLRSlot() for s in used_slots):
    if config.get('enable_hbm_binding_adjustment', False):
      config['new_hbm_binding'] = {}
      left_curr = 0
      right_curr = 16
      for v, s in v2s.items():
        if  config['vertices'][v.name]['category'] == 'PORT_VERTEX' and \
            config['vertices'][v.name]['port_cat'] == 'HBM':
          assert s.getSLR() == 0
          name = config['vertices'][v.name]['top_arg_name']
          if s.isLeftHalf():
            port = left_curr
            left_curr += 1
          elif s.isRightHalf():
            port = right_curr
            right_curr += 1
          else:
            assert False

          config['new_hbm_binding'][name] = port


  return config



import logging
from typing import Dict, List, Tuple

from autobridge.Floorplan.Utilities import (
  log_resource_utilization,
  get_actual_area_limit,
  get_actual_slr_crossing_limit,
)
from autobridge.Floorplan.EightWayPartition import eight_way_partition
from autobridge.Floorplan.FourWayPartition import four_way_partition
from autobridge.Opt.DataflowGraph import Vertex
from autobridge.Opt.Slot import Slot
from autobridge.Opt.SlotManager import SlotManager

_logger = logging.getLogger('autobridge')


def partition(
  init_v2s: Dict[Vertex, Slot],
  slot_manager: SlotManager,
  grouping_constraints: List[List[Vertex]],
  pre_assignments: Dict[Vertex, Slot],
  min_area_limit: float = 0.65,
  max_area_limit: float = 0.85,
  min_slr_width_limit: int = 10000,
  max_slr_width_limit: int = 15000,
  max_search_time: int = 600,
  partition_method: str = 'EIGHT_WAY_PARTITION',
  floorplan_opt_priority: str = 'AREA_PRIORITIZED',
  hbm_port_v_list: List[Vertex] = [],
) -> Dict[Vertex, Slot]:
  _logger.info('Use partition method: %s with floorplan_opt_priority %s', partition_method, floorplan_opt_priority)

  if partition_method == 'EIGHT_WAY_PARTITION':
    partitioner = eight_way_partition
  elif  partition_method == 'FOUR_WAY_PARTITION':
    partitioner = four_way_partition
  else:
    raise NotImplementedError('unrecognized floorplan method: %s', partition_method)

  if floorplan_opt_priority == 'AREA_PRIORITIZED':
    worker = partition_area_prioritized
  elif floorplan_opt_priority == 'SLR_CROSSING_PRIORITIZED':
    worker = partition_slr_crossing_prioritized
  else:
    raise NotImplementedError('unrecognized floorplan_opt_priority: %s', floorplan_opt_priority)

  v2s = worker(
    init_v2s,
    slot_manager,
    grouping_constraints,
    pre_assignments,
    min_area_limit,
    max_area_limit,
    min_slr_width_limit,
    max_slr_width_limit,
    max_search_time,
    hbm_port_v_list,
    partitioner,
  )

  if v2s:
    log_resource_utilization(v2s)

  return v2s


def partition_area_prioritized(
  init_v2s: Dict[Vertex, Slot],
  slot_manager: SlotManager,
  grouping_constraints: List[List[Vertex]],
  pre_assignments: Dict[Vertex, Slot],
  min_area_limit: float,
  max_area_limit: float,
  min_slr_width_limit: int,
  max_slr_width_limit: int,
  max_search_time: int,
  hbm_port_v_list: List[Vertex],
  partitioner,
) -> Dict[Vertex, Slot]:
  """
  adjust the max_usage_ratio if failed
  """
  # use the max slr limit to find the min area limit
  curr_v2s, curr_area_limit = _binary_search_area_limit(
    init_v2s,
    slot_manager,
    grouping_constraints,
    pre_assignments,
    max_slr_width_limit,
    min_area_limit,
    max_area_limit,
    max_search_time,
    hbm_port_v_list,
    partitioner,
  )

  if not curr_v2s:
    _logger.info(f'partition failed with max_usage_ratio {max_area_limit} and slr_width_limit {max_slr_width_limit}')
    return {}

  # use this value as the upper bound in binary search
  actual_slr_width_usage = get_actual_slr_crossing_limit(curr_v2s)

  _logger.info(f'partition succeeds with max_usage_ratio {curr_area_limit} and slr_width_limit {actual_slr_width_usage}')
  return curr_v2s


def partition_slr_crossing_prioritized(
  init_v2s: Dict[Vertex, Slot],
  slot_manager: SlotManager,
  grouping_constraints: List[List[Vertex]],
  pre_assignments: Dict[Vertex, Slot],
  min_area_limit: float,
  max_area_limit: float,
  min_slr_width_limit: int,
  max_slr_width_limit: int,
  max_search_time: int,
  hbm_port_v_list: List[Vertex],
  partitioner,
) -> Dict[Vertex, Slot]:
  """
  adjust the max_usage_ratio if failed
  """
  # use the max area to find the min slr limit
  curr_v2s, curr_slr_limit = _binary_search_slr_crossing_limit(
    init_v2s,
    slot_manager,
    grouping_constraints,
    pre_assignments,
    max_area_limit,
    min_slr_width_limit,
    max_slr_width_limit,
    max_search_time,
    hbm_port_v_list,
    partitioner,
  )

  if not curr_v2s:
    _logger.info(f'partition failed with max_usage_ratio {max_area_limit} and slr_width_limit {max_slr_width_limit}')
    return {}

  # use this value as the upper bound in binary search
  actual_area_usage = get_actual_area_limit(curr_v2s)

  _logger.info(f'partition succeeds with max_usage_ratio {actual_area_usage} and slr_width_limit {curr_slr_limit}')
  return curr_v2s


def _binary_search_slr_crossing_limit(
  init_v2s: Dict[Vertex, Slot],
  slot_manager: SlotManager,
  grouping_constraints: List[List[Vertex]],
  pre_assignments: Dict[Vertex, Slot],
  area_limit: float,
  min_slr_width_limit: int,
  max_slr_width_limit: int,
  max_search_time: int,
  hbm_port_v_list: List[Vertex],
  partitioner,
) -> Tuple[Dict[Vertex, Slot], int]:
  _logger.info('Binary search of slr_width_limit between %d and %d, '
               'with area_limit %f', min_slr_width_limit, max_slr_width_limit, area_limit)

  curr_best_v2s = {}

  # binary search:
  hi = max_slr_width_limit
  lo = min_slr_width_limit

  curr_min_slr_limit = None
  while (lo <= hi):
    curr_slr_limit = (hi + lo) / 2
    _logger.info(f'Try slr_width_limit {curr_slr_limit}')

    v2s = partitioner(
      init_v2s,
      grouping_constraints,
      pre_assignments,
      slot_manager,
      area_limit,
      curr_slr_limit,
      max_search_time,
      hbm_port_v_list=hbm_port_v_list,
    )

    if v2s:
      curr_best_v2s = v2s

      # the actual area usage may be smaller than the provided limit
      actual_slr_usage = get_actual_slr_crossing_limit(v2s)
      curr_min_slr_limit = actual_slr_usage
      hi = actual_slr_usage

      _logger.info('succeed with max actual slr crossing count %d', actual_slr_usage)
    else:
      lo = curr_slr_limit

    if hi - lo < 500:
      break

  if curr_best_v2s:
    _logger.info(f'Found solution with area limit {area_limit} and slr_width_limit {curr_min_slr_limit}')

  return curr_best_v2s, curr_min_slr_limit


def _binary_search_area_limit(
  init_v2s: Dict[Vertex, Slot],
  slot_manager: SlotManager,
  grouping_constraints: List[List[Vertex]],
  pre_assignments: Dict[Vertex, Slot],
  slr_width_limit,
  min_area_limit,
  max_area_limit,
  max_search_time,
  hbm_port_v_list: List[Vertex],
  partitioner,
) -> Tuple[Dict[Vertex, Slot], float]:
  _logger.info('Binary search of min_area_limit between %f and %f, '
               'with slr_width_limit %d', min_area_limit, max_area_limit, slr_width_limit)

  curr_best_v2s = {}

  # binary search:
  hi = max_area_limit
  lo = min_area_limit

  curr_min_usage = None
  while (lo <= hi):
    curr_area_limit = (hi + lo) / 2
    _logger.info(f'Try area limit {curr_area_limit}')

    v2s = partitioner(
      init_v2s,
      grouping_constraints,
      pre_assignments,
      slot_manager,
      curr_area_limit,
      slr_width_limit,
      max_search_time,
      hbm_port_v_list=hbm_port_v_list,
    )

    if v2s:
      curr_best_v2s = v2s

      # the actual area usage may be smaller than the provided limit
      actual_area_usage = get_actual_area_limit(v2s)
      curr_min_usage = actual_area_usage
      hi = actual_area_usage

      _logger.info('succeed with actual area usage %f', actual_area_usage)
    else:
      lo = curr_area_limit

    if hi - lo < 0.01:
      break

  if curr_best_v2s:
    _logger.info(f'Found solution with usage_ratio {curr_min_usage} and slr_width_limit {slr_width_limit}')

  return curr_best_v2s, curr_min_usage



import logging
from typing import List, Dict

from autobridge.util import get_mip_model_silent
from autobridge.Opt.DataflowGraph import Edge, Vertex
from autobridge.Opt.Slot import Slot
from autobridge.util import get_cli_logger
from mip import INTEGER, minimize, xsum, OptimizationStatus

_logger = logging.getLogger('autobridge')
cli_logger = get_cli_logger()


def get_latency(path: List[Slot]) -> int:
  """[TODO] This function should synchronize with the codegen"""
  return len(path)


def latency_balancing(graph, fifo_to_path) -> Dict[str, int]:
    name_to_edge: Dict[str, Edge] = graph.getNameToEdgeMap()
    name_to_vertex: Dict[str, Vertex] = graph.getNameToVertexMap()

    m = get_mip_model_silent()

    # map Vertex -> "arrival time"
    vertex_to_var = {}
    for name, v in name_to_vertex.items():
      vertex_to_var[v] = m.add_var(var_type=INTEGER, name=name)

    # differential constraint for each edge
    for e_name, e in name_to_edge.items():
      # +1 because each FIFO by itself has 1 unit of latency
      # in case the original design is not balanced
      # note that additional pipelining for full_n will not lead to additional latency
      # we only need to increase the grace period of almost full FIFOs by 1
      # [update]: we skip the +1 for the orginial FIFO. We only take care of our own modifications
      m += vertex_to_var[e.src] >= vertex_to_var[e.dst] + get_latency(fifo_to_path[e])

    m.objective = minimize(xsum(
      e.width * (vertex_to_var[e.src] - vertex_to_var[e.dst]) for e in name_to_edge.values()
    ))

    status = m.optimize(max_seconds=120)
    if status != OptimizationStatus.OPTIMAL and status != OptimizationStatus.FEASIBLE:
      cli_logger.warning('Failed to balance reconvergent paths at loop level. Most likely there is a loop of streams.')
      return {}

    # get result
    fifo_name_to_depth = {}
    for e_name, e in name_to_edge.items():
      e.added_depth_for_rebalance = int(vertex_to_var[e.src].x - vertex_to_var[e.dst].x) - e.pipeline_level
      fifo_name_to_depth[e_name] = e.depth + e.added_depth_for_rebalance
      assert e.added_depth_for_rebalance >= 0

    # logging
    for e_name, e in name_to_edge.items():
      _logger.info(f'{e_name}: pipeline_level: {e.pipeline_level}, original depth: {e.depth}, added_depth_for_rebalance: {e.added_depth_for_rebalance}, width: {e.width} ')

    return fifo_name_to_depth
